/**
 * @file fbt_syscall.c
 * Implementation of thread local trampolines.
 *
 * Copyright (c) 2011 ETH Zurich
 * @author Mathias Payer <mathias.payer@nebelwelt.net>
 *
 * $Date: 2011-03-23 10:26:53 +0100 (Wed, 23 Mar 2011) $
 * $LastChangedDate: 2011-03-23 10:26:53 +0100 (Wed, 23 Mar 2011) $
 * $LastChangedBy: payerm $
 * $Revision: 443 $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */
#include "fbt_syscall.h"

#include <assert.h>

#include "fbt_code_cache.h"
#include "fbt_datatypes.h"
#include "fbt_debug.h"
#include "fbt_libc.h"
#include "fbt_llio.h"
#include "fbt_mem_mgmt.h"
#include "fbt_translate.h"
#include "libfastbt.h"

#if defined(AUTHORIZE_SYSCALLS)
/*
 * System call authorization functions must ensure a couple of things:
 * - First of all they may not change or write any of their arguments
 * - Second they must verify that the syscall_nr is correct (even if they assume
 *   that they are only called from one syscall number)
 *
 * These functions are called by asm-magic and include all possible
 * parameters. If you change any parameters then they will be propagated back
 * and the changed parameters will be used for the system call if it is allowed.
 */

/**
 * Prints all arguments of any registered system call (think about ptrace)
 * @return Allows the current system call
 */
static enum syscall_auth_response debug_syscall(struct thread_local_data *tld,
                                                ulong_t syscall_nr, ulong_t arg1,
                                                ulong_t arg2, ulong_t arg3,
                                                ulong_t arg4, ulong_t arg5,
                                                ulong_t *arg6,
                                                ulong_t is_sysenter,
                                                ulong_t *retval);

/**
 * Denies the current system call.
 * @return Denies the current system call and returns -1
 */
static enum syscall_auth_response deny_syscall(struct thread_local_data *tld,
                                               ulong_t syscall_nr, ulong_t arg1,
                                               ulong_t arg2, ulong_t arg3,
                                               ulong_t arg4, ulong_t arg5,
                                               ulong_t *arg6,
                                               ulong_t is_sysenter,
                                               ulong_t *retval);

/**
 * Allows the current system call.
 * @return Allows the current system call and passes control to the kernel.
 */
static enum syscall_auth_response allow_syscall(struct thread_local_data *tld,
                                                ulong_t syscall_nr, ulong_t arg1,
                                                ulong_t arg2, ulong_t arg3,
                                                ulong_t arg4, ulong_t arg5,
                                                ulong_t *arg6,
                                                ulong_t is_sysenter,
                                                ulong_t *retval);

/**
 * Catches an execve system call and checks the parameters. If the parameters
 * still contain the LD_PRELOAD of our BT then the system call is allowed.
 * @return Allows the system call if parameters are OK.
 */
static enum syscall_auth_response auth_execve(struct thread_local_data *tld,
                                              ulong_t syscall_nr, ulong_t arg1,
                                              ulong_t arg2, ulong_t arg3,
                                              ulong_t arg4, ulong_t arg5,
                                              ulong_t *arg6,
                                              ulong_t is_sysenter,
                                              ulong_t *retval);

/**
 * Checks the parameters of an mmap and ensures that the region does not overlap
 * with any BT region. It also checks if new code is marked executable.
 * @return Allows the system call if parameters are OK.
 */
static enum syscall_auth_response auth_mmap(struct thread_local_data *tld,
                                            ulong_t syscall_nr, ulong_t arg1,
                                            ulong_t arg2, ulong_t arg3,
                                            ulong_t arg4, ulong_t arg5,
                                            ulong_t *arg6,
                                            ulong_t is_sysenter,
                                            ulong_t *retval);

/**
 * Checks the parameters of an mprotect and ensures that the application does
 * not have access to any BT region. It also checks if new code is marked
 * executable.
 * @return Allows the system call if parameters are OK.
 */
static enum syscall_auth_response auth_mprotect(struct thread_local_data *tld,
                                                ulong_t syscall_nr, ulong_t arg1,
                                                ulong_t arg2, ulong_t arg3,
                                                ulong_t arg4, ulong_t arg5,
                                                ulong_t *arg6,
                                                ulong_t is_sysenter,
                                                ulong_t *retval);

#if defined(HANDLE_SIGNALS)
/**
 * Checks the parameters of a signal system call and verifies that the signal
 * handler will be instrumented with a trampoline if the signal is delivered.
 * @return Allows the system call if parameters are OK.
 */
static enum syscall_auth_response auth_signal(struct thread_local_data *tld,
                                              ulong_t syscall_nr, ulong_t arg1,
                                              ulong_t arg2, ulong_t arg3,
                                              ulong_t arg4, ulong_t arg5,
                                              ulong_t *arg6,
                                              ulong_t is_sysenter,
                                              ulong_t *retval);

/**
 * Special handler for illegal signals. If such a signal is received then the BT
 * and the application are shut down.
 */
static void illegal_signal();

/**
 * Special handler for ignored signals. Just returns without changeing anything.
 */
static void ignore_signal();

/**
 * Initialization function for the thread-local signal handlers.
 * @param tld Pointer to thread local data.
 */
static void init_signal_handlers(struct thread_local_data *tld);

/**
 * Installs a new signal handler for a specific signal.
 * @param tld Pointer to thread local data.
 * @param signal signal number
 * @param function pointer to signal handler. 
 */
static void install_signal(struct thread_local_data *tld, long signal,
                           void *function);

/**
 * Special handler for illegal signals that prints a warning message and goes
 * into a spin loop.
 */
static void seghandler();
#endif  /* HANDLE_SIGNALS */

#if defined(HANDLE_THREADS)
static enum syscall_auth_response auth_clone(struct thread_local_data *tld,
                                             ulong_t syscall_nr, ulong_t arg1,
                                             ulong_t arg2, ulong_t arg3,
                                             ulong_t arg4, ulong_t arg5,
                                             ulong_t *arg6,
                                             ulong_t is_sysenter,
                                             ulong_t *retval);

static enum syscall_auth_response auth_exit(struct thread_local_data *tld,
                                            ulong_t syscall_nr, ulong_t arg1,
                                            ulong_t arg2, ulong_t arg3,
                                            ulong_t arg4, ulong_t arg5,
                                            ulong_t *arg6,
                                            ulong_t is_sysenter,
                                            ulong_t *retval);
#endif  /* HANDLE_THREADS */

#if defined(HANDLE_SIGNALS)
static void illegal_signal() {
  fbt_suicide_str("Illegal signal received (fbt_syscall.c).\n");
}

static void ignore_signal() {
}

static void init_signal_handlers(struct thread_local_data *tld) {
  long i;
  for (i = 0; i < MAX_NR_SIGNALS; ++i) {
    tld->signal_handlers[i] = &illegal_signal;
  }

  long retval;
  fbt_signal(4, &seghandler, retval);  /* SIGILL */
  fbt_signal(7, &seghandler, retval);  /* SIGBUS */
  fbt_signal(11, &seghandler, retval);  /* SIGSEGV */
}

static void seghandler() {
  fllwrite(STDOUT_FILENO, "Something bad happened. Attach a debugger NOW.\n");
  while (1);
}

static void install_signal(struct thread_local_data *tld, long signal,
                           void *function) {
  if (signal < 0 || signal >= MAX_NR_SIGNALS) {
    fbt_suicide_str("New handler for illegal signal number (fbt_syscall.c)\n");
  }

  /* save old function pointer for later */
  void *might_be_trampoline = tld->signal_handlers[signal];

  /* create a trampoline for this (new) function */
  tld->signal_handlers[signal] = (void (*)(long))       \
    fbt_create_trampoline(tld, function, &(tld->signal_handlers[signal]),
                          ORIGIN_ABSOLUTE);

  /* check if the current handler points to a trampoline, if so then free that
     trmapoline */
  struct mem_info *mem_info = tld->chunk;
  while (mem_info != NULL) {
    if (PTR_IN_REGION(might_be_trampoline, mem_info->ptr, mem_info->size)) {
      /* different types of internal memory that could match */
      switch (mem_info->type) {
      case MT_TRAMPOLINE:
        /* free this trampoline again */
        PRINT_DEBUG("Releasing old trampoline (%p) for signal handler\n",
                    might_be_trampoline);
        fbt_trampoline_free(tld, (struct trampoline*)might_be_trampoline);
        break;
      case MT_CODE_CACHE:
        PRINT_DEBUG("The code of the function for this signal has already been" \
                    " translated.\n");
        break;
      default:
        fbt_suicide_str("Signal handler points into internal BT data " \
                        "(fbt_syscall.c)\n");
      }
    }
    mem_info = mem_info->next;
  }
}

static enum syscall_auth_response auth_signal(struct thread_local_data *tld,
                                              ulong_t syscall_nr, ulong_t arg1,
                                              ulong_t arg2, ulong_t arg3,
                                              ulong_t arg4, ulong_t arg5,
                                              ulong_t *arg6,
                                              ulong_t is_sysenter,
                                              ulong_t *retval) {
  if (syscall_nr != SYS_signal && syscall_nr != SYS_sigaction &&
      syscall_nr != SYS_rt_sigaction) {
    fbt_suicide_str("Invalid system call number in signal auth (fbt_syscall.c).");
  }

  /* arg1: signal number
     arg2: { const struct sigaction *act | sighandler_t }
   */
  if (syscall_nr == SYS_signal) {
    void *function = NULL;
    switch (arg2) {
    case SIG_DFL:
      function = &illegal_signal;
      break;
    case SIG_IGN:
      function = &ignore_signal;
      break;
    default:
      function = (void*)arg2;
    }
    install_signal(tld, arg1, function);
    fbt_signal(arg1, arg2, retval);
    return SYSCALL_AUTH_FAKE;
  }
  
  if (syscall_nr == SYS_sigaction || syscall_nr == SYS_rt_sigaction) {
    /* interpret the sigaction struct */
    ulong_t *sigaction = (ulong_t*)arg2;
    void *function = NULL;
    switch (sigaction[0]) {
    case SIG_DFL:
      function = &illegal_signal;
      break;
    case SIG_IGN:
      function = &ignore_signal;
      break;
    default:
      function = (void*)(sigaction[0]);
    }
    install_signal(tld, arg1, function);
    /* our private copy of the sigaction struct that is passed to the internal
       implementation of this syscall */
    ulong_t arg[3];
    arg[0] = (ulong_t)(function);
    arg[1] = sigaction[1];
    arg[2] = sigaction[2];
    if (syscall_nr == SYS_sigaction) {
      fbt_sigaction(arg1, &arg, arg3, retval);
    } else {
      fbt_rt_sigaction(arg1, &arg, arg3, retval);
    } 
    return SYSCALL_AUTH_FAKE;
  }

  return deny_syscall(tld, syscall_nr, arg1, arg2, arg3, arg4, arg5, arg6,
                      is_sysenter, retval);
}
#endif  /* HANDLE_SIGNALS */

#if defined(HANDLE_THREADS)
static enum syscall_auth_response auth_clone(struct thread_local_data *tld,
                                             ulong_t syscall_nr, ulong_t arg1,
                                             ulong_t arg2, ulong_t arg3,
                                             ulong_t arg4, ulong_t arg5,
                                             ulong_t *arg6 __attribute__((unused)),
                                             ulong_t is_sysenter,
                                             ulong_t *retval) {
  if (syscall_nr != SYS_clone) {
    fbt_suicide_str("Invalid system call number in clone auth (fbt_syscall.c).");
  }

  if (!(arg1 & CLONE_VM)) {
    long local_ret;
#if defined(DEBUG)
    llprintf("Syscall granted (fork through clone)\n");
    ffflush();
#endif
    __asm__ __volatile__("pushl %%ebx\n"
                         "movl %1, %%ebx\n"
                         "movl %2, %%ecx\n"
                         "movl %3, %%edx\n"
                         "movl %4, %%esi\n"
                         "movl %5, %%edi\n"
                         "movl $0x78, %%eax\n"
                         "int $0x80\n" 
                         "popl %%ebx\n"
                         "movl %%eax, %0\n"
                         : "=rmi"(local_ret)
                         : "rmi"(arg1), "rmi"(arg2), "rmi"(arg3), "rmi"(arg4),
                           "rmi"(arg5)
                         : "memory", "cc", "eax", "ecx", "edx", "esi", "edi"
                         );
    *retval = local_ret;
#if defined(DEBUG)
    if (local_ret != 0) {
      llprintf("New process (pid: %d)\n", local_ret);
      ffflush();
    }
#endif
    return SYSCALL_AUTH_FAKE;  
  }
  
  /* we start a new thread */
  if ((arg1 & CLONE_VM) && !is_sysenter) {
    long local_ret;
    /* jump over that int 0x80 or sysenter instruction (both are 2bytes long) */
    void *syscall_location = (void*)(((ulong_t)tld->syscall_location)+2);

    /* initialize new BT data structures for the new thread */
    struct thread_local_data *new_threads_tld = fbt_init(NULL);
    fbt_ccache_add_entry(new_threads_tld, (void*)fbt_commit_transaction,
                         (void*)fbt_end_transaction);
  
#if defined(HIJACKCONTROL)
    fbt_ccache_add_entry(new_threads_tld, (void*)fbt_exit, (void*)fbt_exit);
#endif  /* HIJACKCONTROL */

    /* translate the TU if not already in tcache */
    ulong_t *childsp = (ulong_t*)(arg2 - sizeof(void*));
    struct trampoline *trampo = fbt_create_trampoline(new_threads_tld,
                                                      syscall_location, childsp,
                                                      ORIGIN_CLEAR);
    /* start translation in child through trampoline */
    *childsp = (ulong_t)trampo;
    
    /* start the new thread (execute system call) */
    /* the stack of the child is invalid after this system call,
       so better fix it! */
    __asm__ __volatile__("pushl %%ebx\n"
                         "movl %2, %%ecx\n"
                         "leal -4(%%ecx), %%ecx\n"
                         "movl %%ebx, (%%ecx)\n"
                         "movl %1, %%ebx\n"
                         "movl %3, %%edx\n"
                         "movl %4, %%esi\n"
                         "movl %5, %%edi\n"
                         "movl $0x78, %%eax\n"
                         "int $0x80\n"
                         "cmpl $0x0, %%eax\n" 
                         "popl %%ebx\n"
                         "jne 1f\n" 
                         /* ok, we are the child, let's bail out */
                         "ret\n"
                         "1:\n"
                         "movl %%eax, %0\n"
                         : "=m"(local_ret)
                         : "m"(arg1), "m"(childsp),
                           "m"(arg3), "m"(arg4), "m"(arg5)
                         : "memory", "cc", "eax", "ecx", "edx", "esi", "edi"
                         );

    /* we are the parent thread, let's return the result from the clone syscall */
    *retval = local_ret;
#if defined(DEBUG)
    llprintf("New thread (pid: %d)\n", local_ret);
    ffflush();
#endif
    return SYSCALL_AUTH_FAKE;
  }

  fbt_suicide_str("Unhandled combination of arguments for clone.\n");
  return SYSCALL_AUTH_FAKE;
  
}

static enum syscall_auth_response auth_exit(struct thread_local_data *tld,
                                            ulong_t syscall_nr, ulong_t arg1,
                                            ulong_t arg2 __attribute__((unused)),
                                            ulong_t arg3 __attribute__((unused)),
                                            ulong_t arg4 __attribute__((unused)),
                                            ulong_t arg5 __attribute__((unused)),
                                            ulong_t *arg6 __attribute__((unused)),
                                            ulong_t is_sysenter
                                            __attribute__((unused)),
                                            ulong_t *retval
                                            __attribute__((unused))) {
  if ((syscall_nr != SYS_exit) && (syscall_nr != SYS_exit_group)) {
    fbt_suicide_str("Invalid system call number in exit auth (fbt_syscall.c).");
  }
  
  /* we are shutting down this thread -> clean up BT */
#if defined(DEBUG)
  llprintf("thread/process exit (%p, retval: %d)\n", tld, arg1);
  ffflush();
#endif
  /* we are in the context of the BT, but we might want to print some
     statistics... (otherwise end_transaction would not be needed)  */
  fbt_end_transaction();
  fbt_exit(tld);

  /* fbt_exit unmaps all memory except the last and final pages for the tld.
     we need this storage because we are currently running on this stack.
     So we need a careful trick to get rid of that last memory. We therefore
     call munmap directly in an assembler sequence. After the munmap call the
     stack is no longer valid, so we need to keep all data that we need after
     that syscall in registers. */
  /* this system call will never return, so don't bother about a clean stack */
  if (syscall_nr == SYS_exit) {
    __asm__ __volatile__("movl %0, %%eax\n"
                         "movl %1, %%ebx\n"
                         "movl %2, %%ecx\n"
                         "int $0x80\n"
                         "movl %3, %%eax\n"
                         "movl %4, %%ebx\n"
                         "int $0x80\n"
                         "hlt\n"
                         : /* no return value */
                         : "i"(SYS_munmap), "m"(tld->chunk->ptr),
                         "m"(tld->chunk->size), "i"(SYS_exit), "r"(arg1)
                         : "memory", "eax", "ecx");
  } else {
    __asm__ __volatile__("movl %0, %%eax\n"
                         "movl %1, %%ebx\n"
                         "movl %2, %%ecx\n"
                         "int $0x80\n"
                         "movl %3, %%eax\n"
                         "movl %4, %%ebx\n"
                         "int $0x80\n"
                         "hlt\n"
                         : /* no return value */
                         : "i"(SYS_munmap), "m"(tld->chunk->ptr),
                         "m"(tld->chunk->size), "i"(SYS_exit_group), "r"(arg1)
                         : "memory", "eax", "ecx");
  }
      
  
  fbt_suicide_str("Failed to exit thread/process (fbt_syscall.c)\n");
  return SYSCALL_AUTH_FAKE;
}
#endif  /* HANDLE_THREADS */

static enum syscall_auth_response __attribute__((unused))
debug_syscall(struct thread_local_data *tld __attribute__((unused)),
              ulong_t syscall_nr, ulong_t arg1, ulong_t arg2, ulong_t arg3,
              ulong_t arg4, ulong_t arg5, ulong_t *arg6, ulong_t is_sysenter,
              ulong_t *retval __attribute__((unused))) {
  if (is_sysenter) {
    llprintf("Syscall: %d (arguments: 0x%x 0x%x 0x%x 0x%x 0x%x, ebp: %p, from "
           "sysenter)\n", syscall_nr, arg1, arg2, arg3, arg4, arg5, arg6);
  } else {
    llprintf("Syscall: %d (arguments: 0x%x 0x%x 0x%x 0x%x 0x%x, ebp: %p, from "
           "int)\n", syscall_nr, arg1, arg2, arg3, arg4, arg5, arg6);
  }
  return SYSCALL_AUTH_GRANTED;
}

static enum syscall_auth_response
deny_syscall(struct thread_local_data *tld __attribute__((unused)),
             ulong_t syscall_nr, ulong_t arg1, ulong_t arg2, ulong_t arg3,
             ulong_t arg4, ulong_t arg5, ulong_t *arg6, ulong_t is_sysenter,
             ulong_t *retval) {
  if (is_sysenter) {
    llprintf("Syscall: %d (arguments: 0x%x 0x%x 0x%x 0x%x 0x%x, ebp: %p, from "
           "sysenter)\n", syscall_nr, arg1, arg2, arg3, arg4, arg5, arg6);
  } else {
    llprintf("Syscall: %d (arguments: 0x%x 0x%x 0x%x 0x%x 0x%x, ebp: %p, from "
           "int)\n", syscall_nr, arg1, arg2, arg3, arg4, arg5, arg6);
  }
  fbt_suicide_str("This system call is illegal (fbt_syscall.c).\n");
  *retval = -1;
  return SYSCALL_AUTH_FAKE;
}

static enum syscall_auth_response
allow_syscall(struct thread_local_data *tld __attribute__((unused)),
              ulong_t syscall_nr __attribute__((unused)),
              ulong_t arg1 __attribute__((unused)),
              ulong_t arg2 __attribute__((unused)),
              ulong_t arg3 __attribute__((unused)),
              ulong_t arg4 __attribute__((unused)),
              ulong_t arg5 __attribute__((unused)),
              ulong_t *arg6 __attribute__((unused)),
              ulong_t is_sysenter __attribute__((unused)),
              ulong_t *retval __attribute__((unused))) {
  return SYSCALL_AUTH_GRANTED;
}

static enum syscall_auth_response
auth_execve(struct thread_local_data *tld __attribute__((unused)),
            ulong_t syscall_nr, ulong_t arg1,
            ulong_t arg2 __attribute__((unused)), ulong_t arg3,
            ulong_t arg4 __attribute__((unused)),
            ulong_t arg5  __attribute__((unused)),
            ulong_t *arg6  __attribute__((unused)),
            ulong_t is_sysenter __attribute__((unused)),
            ulong_t *retval) {
  if (syscall_nr != SYS_execve) {
    fbt_suicide_str("Invalid system call number in execve (fbt_syscall.c).");
  }
  char **envp = (char**)arg3;
  int result = SYSCALL_AUTH_FAKE;
  int i = 0;
  char *env_var;
  for (i = 0; *envp != NULL; ++i, ++envp) {
    env_var = *envp;
    if (fbt_strncmp(env_var, "LD_PRELOAD=", 11) == 0) {
      /* there is an LD_PRELOAD in the environment variables array */
      if (fbt_strncmp(env_var+11, LIBNAME, LIBLEN) == 0) {
        /* LD_PRELOAD has not been modified */
        result = SYSCALL_AUTH_GRANTED;
      } else {
        result = SYSCALL_AUTH_FAKE;
      }
    }
  }
  if (result == SYSCALL_AUTH_FAKE) {
    *retval = -1;
    llprintf("command %s not executed due to modified LD_PRELOAD\n", arg1);
    PRINT_DEBUG("command %s not executed due to modified LD_PRELOAD\n", arg1);
  } else {
    PRINT_DEBUG("command %s executed with LD_PRELOAD=%s\n", arg1, (env_var+11));
  }
  ffflush();
  return result;
}

static enum syscall_auth_response
auth_mmap(struct thread_local_data *tld __attribute__((unused)),
          ulong_t syscall_nr, ulong_t arg1, ulong_t arg2,
          ulong_t arg3 __attribute__((unused)),
          ulong_t arg4 __attribute__((unused)),
          ulong_t arg5 __attribute__((unused)),
          ulong_t *arg6 __attribute__((unused)),
          ulong_t is_sysenter __attribute__((unused)),
          ulong_t *retval __attribute__((unused))) {
  /* In this function mmap and mmap2 are treated equally, so the last parameter
     must be checked independently. */
  if (syscall_nr != SYS_mmap && syscall_nr != SYS_mmap2) {
    fbt_suicide_str("Invalid system call number in mmap (fbt_syscall.c).");    
  }
#ifdef SECU_ALLOW_RUNTIME_ALLOC
  /* TODO: secu allow runtime code alloc */
  if ((arg3 & PROT_EXEC) && (arg4 & MAP_ANONYMOUS)) {
    /* runtime allocation of executable memory without file backing */
    fbt_mmap((void*) arg1, arg2, arg3, arg4 & (~MAP_FIXED), arg5, arg6, retval);
    fbt_memprotect_add_valid((void*) retval, arg2);
    return SYSCALL_AUTH_FAKE;
  }
#endif /* SECU_ALLOW_RUNTIME_ALLOC */

  /* TODO: add check for regions of elf files */

  /* ensure we don't remap memory structures of the BT */
  struct mem_info *mem_info = tld->chunk;
  void *startptr = (void*)arg1;
  void *endptr = (void*)(arg1 + arg2 - 1);
  while (mem_info != NULL) {
    if (PTR_IN_REGION(startptr, mem_info->ptr, mem_info->size) ||
        PTR_IN_REGION(endptr, mem_info->ptr, mem_info->size)) {
      PRINT_DEBUG("Application got access to internal data and tries to mmap" \
                  " our memory. Access rejected. Address: %p, length: %d\n" \
                  "Mem_info: %p, length: %d\n", (void*)arg1, arg2,
                  mem_info->ptr, mem_info->size);
      ffflush();
      fbt_suicide_str("Application tried to mmap internal BT data! " \
                      "(fbt_syscall.c)\n");
    }
    mem_info = mem_info->next;
  }

  return SYSCALL_AUTH_GRANTED;
}

static enum syscall_auth_response
auth_mprotect(struct thread_local_data *tld, ulong_t syscall_nr,
              ulong_t arg1, ulong_t arg2,
              ulong_t arg3 __attribute__((unused)),
              ulong_t arg4 __attribute__((unused)),
              ulong_t arg5 __attribute__((unused)),
              ulong_t *arg6 __attribute__((unused)),
              ulong_t is_sysenter __attribute__((unused)),
              ulong_t *retval __attribute__((unused))) {
  if (syscall_nr != SYS_mprotect) {
    fbt_suicide_str("Invalid system call number in mprotect (fbt_syscall.c).");    
  }

  /* ensure we don't make memory structures of BT executable */
  struct mem_info *mem_info = tld->chunk;
  void *startptr = (void*)arg1;
  void *endptr = (void*)(arg1 + arg2 - 1);
  while (mem_info != NULL) {
    if (PTR_IN_REGION(startptr, mem_info->ptr, mem_info->size) ||
        PTR_IN_REGION(endptr, mem_info->ptr, mem_info->size)) {
      PRINT_DEBUG("Application got access to internal data and tries to mprotect" \
                  " our memory. Access rejected. Address: %p, length: %d\n",
                  (void*)arg1, arg2);
      fbt_suicide_str("Application tried to remap internal BT data! " \
                      "(fbt_syscall.c)\n");
    }
    mem_info = mem_info->next;
  }

  /* TODO: add check for regions of elf files */
  
#ifdef SECU_ALLOW_RUNTIME_ALLOC
  /* TODO: secu allow runtime code alloc */
    if (arg3 & PROT_EXEC) {
        fbt_memprotect_add_valid((void*) arg1, arg2);
        INFO_LLPRINTF("intercepted mprotect call that sets memory executable\n");
    }
#endif /* SECU_ALLOW_RUNTIME_ALLOC */

    return SYSCALL_AUTH_GRANTED;
}

void fbt_init_syscalls(struct thread_local_data *tld) {
  ulong_t i;
  PRINT_DEBUG("Syscall table: %p %p\n", tld->syscall_table, debug_syscall);
  for (i = 0; i <= I386_NR_SYSCALLS; ++i) {
    /* allow_syscall for production, debug_syscall for testing */
    tld->syscall_table[i] = &allow_syscall;
  }
  for (; i < MAX_SYSCALLS_TABLE; ++i) {
    tld->syscall_table[i] = &deny_syscall;    
  }
  /* deny a couple of system calls */
  tld->syscall_table[SYS_ptrace] = &deny_syscall;
  tld->syscall_table[SYS_sigreturn] = &deny_syscall;
  tld->syscall_table[SYS_rt_sigreturn] = &deny_syscall;
  tld->syscall_table[SYS_unused1] = &deny_syscall;
  tld->syscall_table[SYS_unused2] = &deny_syscall;
  tld->syscall_table[SYS_unused3] = &deny_syscall;
  tld->syscall_table[SYS_sys_setaltroot] = &deny_syscall;
  
  /* special handling for special system calls */
  tld->syscall_table[SYS_execve] = &auth_execve;
  tld->syscall_table[SYS_mmap] = &auth_mmap;
  tld->syscall_table[SYS_mmap2] = &auth_mmap;
  tld->syscall_table[SYS_mprotect] = &auth_mprotect;

#if defined(HANDLE_SIGNALS)
  /* redirect system calls that change the system call handlers to our
     validation functions */
  tld->syscall_table[SYS_signal] = &auth_signal;
  tld->syscall_table[SYS_sigaction] = &auth_signal;
  tld->syscall_table[SYS_rt_sigaction] = &auth_signal;
  init_signal_handlers(tld);
#endif  /* HANDLE_SIGNALS */
#if defined(HANDLE_THREADS)
  tld->syscall_table[SYS_clone] = &auth_clone;
  tld->syscall_table[SYS_exit] = &auth_exit;
  tld->syscall_table[SYS_exit_group] = &auth_exit;
#endif  /* HANDLE_THREADS */
}

/**
 * overwrite dlclose so that a libraries can never be unmapped.
 * If a library gets unmapped and a new library gets mapped to the same place then
 * we might execute stale code that is still in the code cache.
 * An other option to forbidding the dlclose is to flush the code cache and the
 * hashtable.
 * TODO: This could be future work
 *
int dlclose(void *handle) {
  llprintf("fbt_syscall.c: dlclose intercepted!\n");
  ffflush();
  return 0;
}*/

#endif  /* AUTHORIZE_SYSCALLS */
