/**
 * @file fbt_dso.h
 * Interface to the shared library loader
 *
 * Copyright (c) 2011 ETH Zurich
 * @author Mathias Payer <mathias.payer@nebelwelt.net>
 *
 * $Date: 2011-03-18 19:09:01 +0100 (Fri, 18 Mar 2011) $
 * $LastChangedDate: 2011-03-18 19:09:01 +0100 (Fri, 18 Mar 2011) $
 * $LastChangedBy: payerm $
 * $Revision: 428 $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */
#ifndef FBT_DSO_H
#define FBT_DSO_H

/* forward declare struct */
struct thread_local_data;


/**
 * This enum defines the possible types of a symbol in a DSO
 */
enum sh_sym_type {
  STT_LOCAL,  /* symbol locally defined in the current DSO */
  STT_REF  /* symbol referenced in another DSO (pointer) */
};

/**
 * Defines a single symbol that is part of the DSO chain.
 * The symbol array per DSO contains the following features.
 */
struct sh_symbol
{
  char *start;  /* start pointer of the symbol */
  long size;  /* size / length of the symbol */
  enum sh_sym_type type;  /* type of the symbol */
};

/**
 * This struct is used to define all loaded DSOs that are scanned by libdetox.
 * The linked list is stored in the thread local data and filled during the scan
 * of the DSOs (triggered whenever there is an unknown symbol).
 * TODO: could this be a security bug? (exploit changes data structures of the
 * ld-linux loader and exploits the fact that we call dl_iterate_phdr, this
 * moves dl_iterate_phdr into the trusted computing base).
 */
struct dso_chain
{
  const char *name;  /**< name of the DSO */
  char *baseaddr;  /**< base address of the object */
  char *endaddr;  /**< end address of the object */
  char *plt;  /**< pointer to the procedure linkage table */
  long plt_size;  /**< size of the plt */
  char *gotplt;  /**< pointer to the global offset table for plts */
  long gotplt_size;  /**< sife of the gotplt */
  char *got;  /**< pointer to the global offset table */
  long got_size;  /**< size of the got */
  struct sh_symbol *symbols;  /**< array of symbols defined in this object */
  long nr_symbols;  /**< size of the array */
  struct dso_chain *next;  /**< next DSO in chain or NULL */
};

/**
 * Defines the type of control transfer between two code locations.
 * These types are then used in \ref fbt_actions.c and the emitted code to
 * verify individual transfers. The checker can then decide on this information
 * and according to the transition rules if this transfer is allowed or not.
 */
enum cftx_type {
  CFTX_CALL,  /**< call instruction (relative to IP) */
  CFTX_JMP,  /**< jmp instruction (relative to IP) */
  CFTX_JMP_IND,  /**< indirect jmp instruction through memory pointer */
  CFTX_CALL_IND,  /**< indirect call instruction through memory pointer */
  CFTX_RET  /**< return instruction, control transfer to eip on stack */
};

/**
 * This function checks a specific control transfer from src to dst.
 * This function is called from different translation functions (static control
 * flow transfers) in \ref fbt_actions.c and embedded into the emitted code (for
 * dynamic control flow transfers).
 * @param tld Thread local data
 * @param src Source address of control transfer
 * @param dst Destination address for control transfer
 * @param type Type of control flow transfer
 */
void fbt_check_transfer(struct thread_local_data *tld, unsigned char *src,
                        unsigned char *dst, enum cftx_type type);

/**
 * Interface to the DSO updating mechanism.
 * This function rescans all loaded DSOs using the dl_iterate_phdr interface of
 * the glibC. All DSOs are checked and the current (thread-local) data-structure
 * is updated according to all loaded DSOs and symbols.
 * @param tld Pointer to thread local data.
 */
void fbt_rescan_dsos(struct thread_local_data *tld);

#endif  /* FBT_DSO_H */
